      SUBROUTINE PBDLACPZ( ICONTXT, UPLO, FORM, DIAG, M, N, A, LDA, B,
     $                     LDB, MINT, NINT, MEN, NEN )
*
*  -- PB-BLAS auxiliary routine (version 2.0) --
*     Oak Ridge National Laboratory and University of Tennessee
*     February 28, 1995
*
*     ..
*     .. Scalar Arguments ..
      CHARACTER*1          UPLO, FORM, DIAG
      INTEGER              ICONTXT, LDA, LDB, M, N, MINT, NINT, MEN, NEN
*     ..
*     .. Array Arguments ..
      DOUBLE PRECISION     A( LDA, * ), B( LDB, * )
*     ..
*
*  Purpose
*  =======
*
*     PBDLACPZ copies part of a two-dimensional upper (or lower) tri-
*     angular Matrix A to another matrix B with forced zeros in the
*     other part.
*
*     ..
*     .. Parameters ..
      DOUBLE PRECISION   ONE,          ZERO
      PARAMETER        ( ONE = 1.0D+0, ZERO = 0.0D+0 )
*     ..
*     .. Local Scalars ..
      LOGICAL            NOUNIT
      INTEGER            I, J, JJ, JP, MN, MX
      DOUBLE PRECISION   DUMMY
*     ..
*     .. External Functions ..
      LOGICAL            LSAME
      INTEGER            ICEIL
      EXTERNAL           LSAME, ICEIL
*     ..
*     .. External Subroutines ..
      EXTERNAL           DCOPY, PBDMATADD, PBDVECADD
*     ..
*     .. Intrinsic Functions ..
      INTRINSIC          MIN
*     ..
*     .. Executable Statements ..
*
      NOUNIT = LSAME( DIAG, 'N' )
      JP = 0
      MN = M
*
      IF( LSAME( UPLO, 'U' ) ) THEN
*
*        A is upper triangular
*
         IF( LSAME( FORM, 'T' ) ) THEN
            DO 20 I = 1, ICEIL( NEN, NINT )
               DO 10 J = 1, MIN( N, NEN-JP )
                  JJ = JP + J
                  MX = MN + J
                  IF( NOUNIT ) THEN
                     CALL DCOPY( MX,   A(1,JJ), 1, B(1,JJ), 1 )
                  ELSE
                     CALL DCOPY( MX-1, A(1,JJ), 1, B(1,JJ), 1 )
                     B(MX,JJ) = ONE
                  END IF
                  CALL PBDVECADD( ICONTXT, 'G', MEN-MX, ZERO, DUMMY, 1,
     $                            ZERO, B(MX+1,JJ), 1 )
   10          CONTINUE
               MN = MN + MINT
               JP = JP + NINT
   20       CONTINUE
*
*        A is a rectangular matrix
*
         ELSE
            DO 30 I = 1, ICEIL( NEN, NINT )
               MX = MIN( N, NEN-JP )
               CALL PBDMATADD( ICONTXT, 'V', MN, MX, ONE, A(1,JP+1),
     $                         LDA, ZERO, B(1,JP+1), LDB )
               CALL PBDMATADD( ICONTXT, 'G', MEN-MN, MX, ZERO, DUMMY, 1,
     $                         ZERO, B(MN+1,JP+1), LDB )
               MN = MN + MINT
               JP = JP + NINT
   30       CONTINUE
         END IF
*
*     If ( LSAME( UPLO, 'L' ) )
*
      ELSE
*
*        A is lower triangular
*
         IF( LSAME( FORM, 'T' ) ) THEN
            MN = M - 1
            DO 50 I = 1, ICEIL( NEN, NINT )
               DO 40 J = 1, MIN( N, NEN-JP )
                  JJ = JP + J
                  MX = MN + J
                  CALL PBDVECADD( ICONTXT, 'G', MX, ZERO, DUMMY, 1,
     $                            ZERO, B(1,JJ), 1 )
                  IF( NOUNIT ) THEN
                     CALL DCOPY( MEN-MX,   A(MX+1,JJ), 1, B(MX+1,JJ),
     $                                 1 )
                  ELSE
                     B(MX+1,JJ) = ONE
                     CALL DCOPY( MEN-MX-1, A(MX+2,JJ), 1, B(MX+2,JJ),
     $                                 1 )
                  END IF
   40          CONTINUE
               MN = MN + MINT
               JP = JP + NINT
   50       CONTINUE
*
*        A is a rectangular matrix
*
         ELSE
            DO 60 I = 1, ICEIL( NEN, NINT )
               MX = MIN( N, NEN-JP )
               CALL PBDMATADD( ICONTXT, 'G', MN, MX, ZERO, DUMMY, 1,
     $                         ZERO, B(1,JP+1), LDB )
               CALL PBDMATADD( ICONTXT, 'V', MEN-MN, MX, ONE,
     $                         A(MN+1,JP+1), LDA, ZERO, B(MN+1,JP+1),
     $                         LDB )
               MN = MN + MINT
               JP = JP + NINT
   60       CONTINUE
         END IF
      END IF
      RETURN
*
*     End of PBDLACPZ
*
      END
